@@ -5,16 +5,18 @@
 # need to read through all of this.  If you're using Postfix and UW POP/IMAP,
 # you can likely just use the default setup without any changes.  The most
 # common changes needed are to pick the right $pat variable for your POP/IMAP
-# software, ensure that the maillog name is right, and perhaps uncomment a
+# software, ensure that the mail-log name is right, and perhaps uncomment a
 # section with the support for a different SMTP (other than Postfix).  See the
 # contrib/README.QUICKSTART file for step-by-step instructions on how to
 # install and test your setup.
 
+use strict;
 use vars qw(
-    $pat $write $flock $debug $reprocess $grace $logto %file_tail
-    @mynets %db $dbfile $dbvalue
-    $mynet_func $tie_func $sync_func $flock_func $log_func
-    $tail_init_func $tail_getline_func
+    $pat $out_pat $write $flock $debug $reprocess $grace $logto %file_tail
+    @mynets %db $dbfile $dbvalue $logtime_pat
+    $mynet_func $tie_func $flock_func $add_func $del_func $sync_func
+    $tail_init_func $tail_getline_func $log_func
+    $PID_pat $IP_pat $OK_pat $FAIL_pat $OUT_pat
 );
 
 #
@@ -111,101 +113,143 @@ sub getline_FileTail
 # an example below).  Feel free to delete all the stuff you don't need.
 #
 # To enable an entry, just delete the "#" at the start of all the lines
-# *after* the initial comment line(s).
+# *after* the initial comment line(s).  If the entry of your choice also
+# provides $out_pat, you should uncomment that variable as well, which
+# allows us to keep track of users who are still connected to the server
+# (e.g. Thunderbird caches open IMAP connections).
 #
 # Note also that the servers that require multiple log lines be read to
-# get all the needed info have a setup with 3 "my" variables instead of
-# a single "$pat" variable.  Uncomment all 3 and we'll define a multi-
+# get all the needed info have a setup with 3 $NAME_pat variables instead
+# of a single "$pat" variable.  Uncomment all 3 and we'll define a multi-
 # line-matching custom_match() subroutine for you.
 #
+# Technical:  if you want to define a new $pat regex, make sure that you
+# include the string "[LOGTIME]" near the start because this string will
+# be replaced by the final value of the $logtime_pat variable, and this
+# handles the matching and returning of the first value the code needs:
+# the line's timestamp.  There must also be a 2nd value returned from the
+# regular expression:  the IP number.
+#
 ############################# START OF PATTERNS #############################
 
-# For UW ipop3d/imapd and their secure versions. This is the DEFAULT.
-#$pat = '^(... .. ..:..:..) \S+ (?:ipop3s?d|imaps?d)\[\d+\]: ' .
+# If your logfile has a non-standard time-stamp, you may need to define the
+# $logtime_pat value here, or use one of the known variations included next
+# to the relevant patterns.  The value should include parentheses, and (for
+# maximum flexibility) should be able to match the time on any log line.
+# This defaults to "(\w\w\w +\d+ \d+:\d+:\d+)" if left undefined.
+
+#$logtime_pat = '(\d\d\d\d-\d\d-\d\d \d+:\d+:\d+)';
+
+# For UW POP/IMAP. This is the DEFAULT $pat (without an $out_pat, by default).
+#$pat = '^[LOGTIME] \S+ (?:ipop3s?d|imaps?d)\[\d+\]: ' .
 #    '(?:Login|Authenticated|Auth|APOP) user=\S+ ' .
 #    'host=(?:\S+ )?\[(\d+\.\d+\.\d+\.\d+)\]';
+#$out_pat = '^[LOGTIME] \S+ (?:ipop3s?d|imaps?d)\[\d+\]: ' .
+#    '(?:Logout|Autologout|Killed.*?|Command stream end of file.*?) user=\S+ host=(?:\S+ )?\[(\d+\.\d+\.\d+\.\d+)\]';
 
 # Some folks need a little different match for UW ipop3d/imapd:
-#$pat = '^(... .. ..:..:..) \S+ (?:ipop3s?d|imaps?d)\[\d+\]: \[[^]]+\]' .
+#$pat = '^[LOGTIME] \S+ (?:ipop3s?d|imaps?d)\[\d+\]: \[[^]]+\]' .
 #    '(?:Login|Authenticated|Auth|APOP) user=\S+ ' .
 #    'host=(?:\S+ )?\[(\d+\.\d+\.\d+\.\d+)\]';
 
 # For GNU pop3d
-#$pat = '^(... .. ..:..:..) \S+ gnu-pop3d\[\d+\]: ' .
-#    'User .+ logged in with mailbox .+ from (\d+\.\d+\.\d+\.\d+)$';
+#$pat = '^[LOGTIME] \S+ gnu-pop3d\[\d+\]: ' .
+#    'User .+ logged in with mailbox .+ from (\d+\.\d+\.\d+\.\d+)';
 
-# A fairly modern Qpopper pattern.
-#$pat = '^(... .. ..:..:..) \S+ (?:in\.)?qpopper[^[]*\[\d+\]: \([^)]*\) ' .
+# A fairly modern Qpopper pattern, when using syslog.
+#$pat = '^[LOGTIME] \S+ (?:in\.q|q)?popper[^[]*\[\d+\]: \([^)]*\) ' .
 #    'POP login by user "[^"]+" at \([^)]+\) (\d+\.\d+\.\d+\.\d+)';
 
+# The same Qpopper pattern as above, but matches Qpopper's own logging.
+#$pat = '^[LOGTIME] \[popper\] \([^)]*\) ' .
+#    'POP login by user (?:"[^"]+"|\S+) at \([^)]+\) (\d+\.\d+\.\d+\.\d+)';
+
 # For Qpopper POP/APOP Server (matches in.qpopper, qpopper, and popper).
-#$pat = '^(... .. ..:..:..) \S+ (?:in\.q|q)?popper\[\d+\]: Stats: \S+ ' .
+#$pat = '^[LOGTIME] \S+ (?:in\.q|q)?popper\[\d+\]: Stats: \S+ ' .
 #    '\d+ \d+ \d+ \d+ (?:\S+ )?(\d+\.\d+\.\d+\.\d+)';
 
 # Chris D.Halverson's pattern for Qpopper 3.0b29 on Solaris 2.6
-#$pat = '^(... ... .. ..:..:.. \d{4}) \[\d+\] ' .
-#    ' Stats:\s+\w+ \d \d \d \d [\w\.]+ (\d+\.\d+\.\d+\.\d+)';
+#$logtime_pat = '(... ... .. \d+:\d+:\d+ \d{4})';
+#$pat = '^[LOGTIME] \[\d+\] Stats:\s+\w+ ' .
+#    '\d \d \d \d [\w\.]+ (\d+\.\d+\.\d+\.\d+)';
 
 # For Cyrus (including a tweak for IP addrs that don't resolve):
-#$pat = '^(... .. ..:..:..) \S+ (?:pop3|imap)[ds]?\[\d+\]: ' .
-#    'login: [^[\s]*\s*\[[:f]*(\d+\.\d+\.\d+\.\d+)\] \S+ \S+';
+#$pat = '^[LOGTIME] \S+ (?:cyrus/)?(?:pop3|imap)[ds]?\[\d+\]: ' .
+#    'login: .*?\[[:f]*(\d+\.\d+\.\d+\.\d+)\]';
 
 # For Courier-POP3 and Courier-IMAP:
-#$pat = '^(... .. ..:..:..) \S+ (?:courier)?(?:pop3|imap)(?:login|d|d-ssl): ' .
+#$pat = '^[LOGTIME] (?:\[|\S+ )(?:pop3|imap|couriertcp)(?:d|d-ssl|login)\]?: ' .
 #    'LOGIN, user=\S+, ip=\[[:f]*(\d+\.\d+\.\d+\.\d+)\]';
+#$out_pat = '^[LOGTIME] (?:\[|\S+ )(?:pop3|imap|couriertcp)(?:d|d-ssl|login)\]?: ' .
+#    '(?:LOGOUT|DISCONNECTED), user=\S+, ip=\[[:f]*(\d+\.\d+\.\d+\.\d+)\]';
 
 # For qmail's pop3d:
-#$pat = '^(... .. ..:..:..) \S+ vpopmail\[\d+\]: ' .
-#    'vchkpw: login \[\S+\] from (\d+\.\d+\.\d+\.\d+)';
+#$pat = '^[LOGTIME] \S+ vpopmail\[\d+\]: vchkpw: ' .
+#    'login \[\S+\] from (\d+\.\d+\.\d+\.\d+)';
 
 # For cucipop, matching a sample from Daniel Roesen:
-#$pat = '^(... .. ..:..:..) \S+ cucipop\[\d+\]: \S+ ' .
+#$pat = '^[LOGTIME] \S+ cucipop\[\d+\]: \S+ ' .
 #    '(\d+\.\d+\.\d+\.\d+) \d+, \d+ \(\d+\), \d+ \(\d+\)';
 
 # A Modern vm-pop3d can use a normal, 1-line pattern:
-#$pat = '^(... .. ..:..:..) \S+ vm-pop3d\[\d+\]: ' .
-#    'User .+ logged in from (\d+\.\d+\.\d+\.\d+)';
+#$pat = '^[LOGTIME] \S+ vm-pop3d\[\d+\]: ' .
+#    'User .+? logged in from (\d+\.\d+\.\d+\.\d+)';
 
-# Older vm-pop3d -- needs to match 2 log entries (uncomment all 3 "my" lines).
-#my $PID_pat = '^(... .. ..:..:..) \S+ vm-pop3d\[(\d+)\]: ';
-#my $IP_pat = $PID_pat . 'Connect from (\d+\.\d+\.\d+\.\d+)$';
-#my $OK_pat = $PID_pat . 'User .+ logged in';
+# Older vm-pop3d -- needs to match 2 log entries (uncomment all 3 *_pat lines).
+#$PID_pat = '^[LOGTIME] \S+ vm-pop3d\[(\d+)\]: ';
+#$IP_pat = $PID_pat . 'Connect from (\d+\.\d+\.\d+\.\d+)';
+#$OK_pat = $PID_pat . 'User .+? logged in';
 
-# For popa3d -- needs to match 2 log entries (uncomment all 3 "my" lines).
-#my $PID_pat = '^(... .. ..:..:..) \S+ popa3d\[(\d+)\]: ';
-#my $IP_pat = $PID_pat . 'Session from (\d+\.\d+\.\d+\.\d+)$';
-#my $OK_pat = $PID_pat . 'Authentication passed for ';
+# For popa3d -- needs to match 2 log entries (uncomment all 3 *_pat lines).
+#$PID_pat = '^[LOGTIME] \S+ popa3d\[(\d+)\]: ';
+#$IP_pat = $PID_pat . 'Session from (\d+\.\d+\.\d+\.\d+)';
+#$OK_pat = $PID_pat . 'Authentication passed for ';
 
 # For *patched* popa3d (see the patch in the contrib/popa3d dir).
-#$pat = '^(... .. ..:..:..) \S+ popa3d\[\d+\]: ' .
+#$pat = '^[LOGTIME] \S+ popa3d\[\d+\]: ' .
 #    'Authentication passed for \S+ -- \[(\d+\.\d+\.\d+\.\d+)\]';
 
 # A Perdition pattern.
-#$pat = '^(... .. ..:..:..) \S+ perdition\[\d+\]: ' .
+#$pat = '^[LOGTIME] \S+ perdition\[\d+\]: ' .
 #    'Auth: (\d+\.\d+\.\d+\.\d+)\-\>\d+\.\d+\.\d+\.\d+ ' .
 #    'user=\"\S+\" server=\"\S+\" port=\"\S+\" status=\"ok\"';
 
-# For solidpop3d (known as spop3d).  ** If you configured
-# solidpop3d with --enable-logextend! **
-#$pat = '^(... .. ..:..:..) \S+ spop3d\[\d+\]: ' .
+# For solidpop3d (aka spop3d) ** if configured with --enable-logextend! **
+#$pat = '^[LOGTIME] \S+ spop3d\[\d+\]: ' .
 #    'user \S+ authenticated - (\d+\.\d+\.\d+\.\d+)';
 
-# For spop3d w/o --enabled-logextend (uncomment all 3 "my" lines).
-#my $PID_pat = '^(... .. ..:..:..) \S+ spop3d\[(\d+)\]: ';
-#my $IP_pat = $PID_pat . 'connect from (\d+\.\d+\.\d+\.\d+)';
-#my $OK_pat = $PID_pat . 'user \S+ authenticated';
+# For spop3d w/o --enabled-logextend (uncomment all 3 *_pat lines).
+#$PID_pat = '^[LOGTIME] \S+ spop3d\[(\d+)\]: ';
+#$IP_pat = $PID_pat . 'connect from (\d+\.\d+\.\d+\.\d+)';
+#$OK_pat = $PID_pat . 'user \S+ authenticated';
 
 # Pattern for teapop (http://www.toontown.org/teapop/) by Patrick Prasse.
-#$pat = '^(... .. ..:..:..) \S+ teapop\[\d+\]: ' .
-#    'Successful login for \S+ .+ \[(\d+\.\d+\.\d+\.\d+)\]$';
+#$pat = '^[LOGTIME] \S+ teapop\[\d+\]: ' .
+#    'Successful login for .+? \[(\d+\.\d+\.\d+\.\d+)\]';
 
-# For Dovecot POP3/IMAP
-#$pat = '^(... .. ..:..:..) \S+ (?:dovecot: )? (?:imap|pop3)-login: ' .
-#    'Login: \S+ \[(\d+\.\d+\.\d+\.\d+)\]';
+# For Dovecot POP3/IMAP when using syslog.
+#$pat = '^[LOGTIME] \S+ (?:dovecot: )?(?:imap|pop3)-login: ' .
+#    'Login: .*? (?:\[|rip=)[:f]*(\d+\.\d+\.\d+\.\d+)[],]';
+#$out_pat = '^[LOGTIME] \S+ (?:dovecot: )?(?:imap|pop3)-login: ' .
+#    'Disconnected.*? (?:\[|rip=)[:f]*(\d+\.\d+\.\d+\.\d+)[],]';
+
+# For Dovecot POP3/IMAP when it does its own logging.
+##$logtime_pat = '(\d\d\d\d-\d+-\d+ \d+:\d+:\d+)';
+#$pat = '^dovecot: [LOGTIME] Info: (?:imap|pop3)-login: ' .
+#    'Login: .+? rip=[:f]*(\d+\.\d+\.\d+\.\d+),';
+#$out_pat = '^dovecot: [LOGTIME] Info: (?:imap|pop3)-login: ' .
+#    'Disconnected.*? rip=[:f]*(\d+\.\d+\.\d+\.\d+),';
+
+# For older Dovecot POP3/IMAP when it does its own logging.
+#$pat = '^(?:imap|pop3)-login: [LOGTIME] Info: ' .
+#    'Login: \S+ \[[:f]*(\d+\.\d+\.\d+\.\d+)\]';
+#$out_pat = '^(?:imap|pop3)-login: [LOGTIME] Info: ' .
+#    'Disconnected.*? \[[:f]*(\d+\.\d+\.\d+\.\d+)\]';
 
 # For Apple IMAP MAIL Server
-#$pat = '^(... .. .... ..:..:..).* IMAP ' .
-#    'User "[^"]+" log on successful from (\d+\.\d+\.\d+\.\d+)';
+#$logtime_pat = '(\w\w\w +\d+ \d\d\d\d \d+:\d+:\d+)';
+#$pat = '^[LOGTIME].* IMAP User "[^"]+" ' .
+#    'log on successful from (\d+\.\d+\.\d+\.\d+)';
 
 ############################## END OF PATTERNS ##############################
 
@@ -215,21 +259,40 @@ sub getline_FileTail
 
 # This is an example of using the custom_match() function to match
 # several patterns (allowing you to match multiple pop/imap servers
-# at the same time).
+# at the same time).  Note that you can define logout patterns for
+# some, all, or none of the login patterns, as needed.
 
-my $pat2 = '... define an extra pattern here ...';
+my(@login, @logout);
 
-# Add as many patterns to the @match array as you like:
-my @match = ( $pat, $pat2 );
+$login[0] = $pat;
+#$logout[0] = $out_pat;
 
-$_ = qr/$_/ foreach @match; # Pre-compile the regular expressions.
+$login[1] = '... define an extra login pattern here ...';
+#$logout[1] = '... define an extra logout pattern, if needed ...';
 
-# The maillog line to match is in $_.
+# ... plus any other patterns you want to define ...
+
+foreach (@login, @logout) {
+    next unless defined $_;
+    $_ = qr/$_/;	# Pre-compile the regular expressions.
+}
+
+# The mail-log line to match is in $_.
 sub custom_match
 {
-    foreach my $regex (@match) {
-	# Return timestamp and IP for any (pre-compiled) pattern that matches.
-	return ($1, $2) if /$regex/;
+    my($timestamp, $ipaddr);
+    foreach my $regex (@logout) {
+	if (defined($regex) && (($timestamp, $ipaddr) = /$regex/)) {
+	    return ($timestamp, $ipaddr, -1);
+	}
+    }
+    my $j = 0;
+    foreach my $regex (@login) {
+	if (($timestamp, $ipaddr) = /$regex/) {
+	    my $increment = defined($logout[$j]) ? 1 : 0;
+	    return ($timestamp, $ipaddr, $increment);
+	}
+	$j++;
     }
     ( );
 }
@@ -246,6 +309,7 @@ sub custom_match
 # If you comment-out (or remove) the two surrounding =cut lines, we'll use
 # NDBM_File instead of DB_File.
 
+use Fcntl;
 use NDBM_File;
 
 #$mynet_func = \&mynet_postfix; # Use the default
@@ -301,7 +365,7 @@ $flock = 0;
 
 sub mynet_tcprules
 {
-    # You'll want to edit this value.
+    # You'll probably want to edit this (it specifies IP ranges to ignore).
     '127.0.0.0/8 192.168.1.1/24';
 }
 
@@ -344,6 +408,7 @@ sub sync_tcprules
 
 my $ESMTPD = '/usr/lib/courier/sbin/esmtpd';
 
+use Fcntl qw(:DEFAULT :flock);
 use DB_File;
 
 $dbfile = '/etc/courier/smtpaccess'; # DB hash to write
@@ -352,7 +417,7 @@ $dbvalue = 'allow,RELAYCLIENT';
 $mynet_func = \&mynet_courier;
 $tie_func = \&tie_courier;
 $sync_func = \&sync_courier;
-#$flock_func = \&flock_DB; # Use the default
+$flock_func = \&flock_courier;
 
 sub mynet_courier
 {
@@ -371,6 +436,13 @@ sub tie_courier
     }
 }
 
+sub flock_courier
+{
+    my($locking) = @_;
+    flock(DB_FH, $locking ? LOCK_EX : LOCK_UN)
+	or die "$0: flock_courier($locking) failed: $!\n";
+}
+
 sub sync_courier
 {
     $dbh->sync and die "$0: sync $dbfile: $!\n" if $write;
@@ -386,6 +458,7 @@ sub sync_courier
 # the sendmail.cf changes you'll need to make.  If you find that Sendmail
 # isn't recognizing the changes to the DB file, set $signal_sendmail to 1.
 
+use Fcntl qw(:DEFAULT :flock :seek);
 use DB_File;
 
 $dbfile = '/etc/mail/popauth'; # DB hash to write
@@ -409,7 +482,7 @@ if ($signal_sendmail) {
 
 sub mynet_sendmail
 {
-    # You'll want to edit this value.
+    # You'll probably want to edit this (it specifies IP ranges to ignore).
     '127.0.0.0/8 192.168.1.1/24';
 }
 
@@ -443,17 +516,20 @@ sub sync_sendmail
     }
 }
 
+# Only 1 of these 2 locking functions is used (look for $flock_func above).
 sub flock_sendmail
 {
-    flock(DB_FH, $_[0]? LOCK_EX : LOCK_UN)
-	or die "$0: flock_DB($_[0]) failed: $!\n";
+    my($locking) = @_;
+    flock(DB_FH, $locking ? LOCK_EX : LOCK_UN)
+	or die "$0: flock_sendmail($locking) failed: $!\n";
 }
 
 sub fcntl_sendmail
 {
-    my $lock = pack('s s l l i', $_[0]? F_WRLCK : F_UNLCK, SEEK_SET, 0, 0, 0);
+    my($locking) = @_;
+    my $lock = pack('s s l l i', $locking ? F_WRLCK : F_UNLCK, SEEK_SET, 0, 0, 0);
     fcntl(DB_FH, F_SETLKW, $lock)
-	or die "$0: flock_DB($_[0]) failed: $!\n";
+	or die "$0: fcntl_sendmail($locking) failed: $!\n";
 }
 =cut #------------------------- Sendmail SMTP ---------------------------END-
 
@@ -492,26 +568,58 @@ if (defined($PID_pat) && defined($IP_pat) && defined($OK_pat)) {
 # Some pop services don't put the IP on the line that lets us know that a
 # user was properly authenticated.  For these programs, we scan the IP off
 # an earlier line and the check the validation by comparing the PID values.
+# The regex $IP_pat is used to match the IP number and cache it.  The regex
+# $OK_pat is used to match a success message that follows from the same PID.
+# If $FAIL_pat is not defined, there can be no intermediate log messages
+# with the same PID prior to the $OK_pat.  If $FAIL_pat is defined, we'll
+# continue trying to match the $OK_pat value on matching PID lines until we
+# either match the $FAIL_pat (in which case the IP is ignored) or the $OK_pat
+# (in which case the IP is accepted).  Thus, $FAIL_pat can be either a
+# failure log line or a this-PID-is-finished log line common to both success
+# and failure.  Finally, an optional $OUT_pat may be defined if you want to
+# track active connections -- when defined, any PID that matches the $OK_pat
+# will be considered connected until we find a match for $OUT_pat.  It is
+# fine if lines that match $OUT_pat would also match $FAIL_pat (such as a
+# disconnect message) -- this won't confuse the code.
+#
+# Keep in mind that $IP_pat, $OK_pat, $OUT_pat, and $FAIL_pat will only be
+# applied to lines that match the $PID_pat regex.
 
-    my %popIPs;
+    my(%popIPs, %popConnected);
 
-    # The maillog line to match is in $_.
+    $FAIL_pat = '.' if !defined $FAIL_pat;
+    foreach ($PID_pat, $IP_pat, $OK_pat, $FAIL_pat, $OUT_pat) {
+	s/\[LOGTIME\]/$logtime_pat/g if defined $_;
+    }
+
+    # The mail-log line to match is in $_.
     sub custom_match
     {
 	if (/$PID_pat/o) {
 	    my($ts, $pid) = ($1, $2);
 	    if (/$IP_pat/o) {
 		$popIPs{$pid} = $3;
-	    }
-	    else {
-		foreach my $key (keys %popIPs) {
-		    if ($pid == $key) {
-			my $ip = $popIPs{$pid};
-			delete $popIPs{$pid};
+	    } else {
+		my $ip = $popIPs{$pid};
+		if (defined $ip) {
+		    if ($popConnected{$pid}) {
+			if (/$OUT_pat/o) {
+			    delete $popConnected{$pid};
+			    delete $popIPs{$pid};
+			    return ($ts, $ip, -1);
+			}
+		    } else {
 			if (/$OK_pat/o) {
-			    return ($ts, $ip);
+			    if (defined $OUT_pat) {
+				$popConnected{$pid} = 1;
+				return ($ts, $ip, 1);
+			    }
+			    delete $popIPs{$pid};
+			    return ($ts, $ip, 0);
+			}
+			if (/$FAIL_pat/o) {
+			    delete $popIPs{$pid};
 			}
-			last;
 		    }
 		}
 	    }
